import contextlib
import json
import re
from pathlib import Path, PurePosixPath

from jailrun.cmd.status.monit import parse_monit_status
from jailrun.cmd.status.types import (
    DiskStats,
    JailRow,
    MemStats,
    MonitJailStatus,
    RawJail,
    StatusInfo,
)
from jailrun.network import SSHKwargs, get_ssh_kw, jail_ssh_exec, ssh_exec, wait_for_ssh
from jailrun.schemas import State
from jailrun.serializers import loads
from jailrun.settings import Settings

RE_IPV4 = re.compile(r"^\s+inet (\d+\.\d+\.\d+\.\d+)", re.MULTILINE)
RE_IPV6 = re.compile(r"^\s+inet6 ([0-9a-f:]+)", re.MULTILINE)
SKIP_V4 = {"127.0.0.1"}
SKIP_V6_PREFIX = ("::1", "fe80:")


def short_path(p: str) -> str:
    parts = Path(p).parts
    if len(parts) > 2:
        return "…/" + str(Path(*parts[-2:]))
    return p


def get_jail_ips(name: str, ssh_kw: SSHKwargs) -> tuple[list[str], list[str]]:
    ipv4 = ipv6 = []
    raw = ssh_exec(cmd=f"jexec {name} ifconfig", ssh_kw=ssh_kw)
    if raw:
        ipv4 = [m for m in RE_IPV4.findall(raw) if m not in SKIP_V4]
        ipv6 = [m for m in RE_IPV6.findall(raw) if not m.startswith(SKIP_V6_PREFIX)]

    return ipv4, ipv6


def get_raw_jails(ssh_kw: SSHKwargs, *, state: State) -> list[RawJail]:
    running: set[str] = set()
    raw = ssh_exec(cmd="jls -N --libxo json", ssh_kw=ssh_kw)
    with contextlib.suppress(json.JSONDecodeError, TypeError):
        for j in loads(raw).get("jail-information", {}).get("jail", []):
            running.add(j["name"])

    datasets = ssh_exec(
        cmd="zfs list -H -o name -d 1 zroot/jailrun/jails",
        ssh_kw=ssh_kw,
    )

    if not datasets:
        return []

    state_by_private = {str(j.private_name): j for j in state.jails.values()}
    names = [PurePosixPath(line).name for line in datasets.strip().splitlines()]
    results: list[RawJail] = []

    for name in names:
        if name == "jails":
            continue

        is_up = name in running

        if is_up:
            ipv4, ipv6 = get_jail_ips(name, ssh_kw)
        else:
            ipv4 = [state_by_private[name].ip or "-" if name in state_by_private else "-"]
            ipv6 = ["-"]

        results.append(RawJail(private_name=name, state="Up" if is_up else "Down", ipv4=ipv4, ipv6=ipv6))

    return results


def get_disk_stats(ssh_kw: SSHKwargs) -> DiskStats:
    disk_free: str | None = None
    disk_total: str | None = None
    disk = ssh_exec(cmd="df -h /", ssh_kw=ssh_kw)
    if disk:
        parts = disk.splitlines()[-1].split()
        if len(parts) >= 4:
            disk_free, disk_total = parts[3], parts[1]

    return DiskStats(disk_free=disk_free, disk_total=disk_total)


def get_mem_stats(ssh_kw: SSHKwargs) -> MemStats:
    mem_total: float | None = None
    mem_usable: float | None = None
    mem = ssh_exec(cmd="sysctl -n hw.physmem hw.usermem", ssh_kw=ssh_kw)
    if mem:
        lines = mem.splitlines()
        if len(lines) == 2:
            mem_total = int(lines[0]) / (1024**3)
            mem_usable = int(lines[1]) / (1024**3)

    return MemStats(mem_total=mem_total, mem_usable=mem_usable)


def _fetch_monit_for_jails(
    jails: list[RawJail],
    public_by_private: dict[str, str],
    ssh_kw: SSHKwargs,
) -> dict[str, MonitJailStatus]:
    monit_by_jail: dict[str, MonitJailStatus] = {}
    running_jails = [j for j in jails if j["state"].lower() == "up"]

    for jail in running_jails:
        jail_ip = " ".join(jail["ipv4"])
        if not jail_ip:
            continue

        raw = jail_ssh_exec(cmd="monit status 2>/dev/null", jail_ip=jail_ip, ssh_kw=ssh_kw, timeout=10)
        if not raw:
            continue

        parsed = parse_monit_status(raw)
        private_name = jail["private_name"]
        public_name = public_by_private.get(private_name, private_name)

        for monit_jail_name, monit_data in parsed.items():
            if monit_jail_name in (private_name, public_name):
                monit_by_jail[public_name] = monit_data
            else:
                entry = monit_by_jail.setdefault(
                    public_name,
                    MonitJailStatus(system_ok=None, services=[]),
                )
                if monit_data["system_ok"] is not None:
                    entry["system_ok"] = monit_data["system_ok"]

                entry["services"].extend(monit_data["services"])

    return monit_by_jail


def collect_info(settings: Settings, state: State, pid: int) -> StatusInfo:
    ssh_kw = get_ssh_kw(settings, state)

    wait_for_ssh(ssh_kw=ssh_kw, silent=True)
    uptime = ssh_exec(cmd="uptime", ssh_kw=ssh_kw)
    disk_stats = get_disk_stats(ssh_kw)
    mem_stats = get_mem_stats(ssh_kw)
    raw_jails = get_raw_jails(ssh_kw, state=state)

    managed_names = set(state.jails.keys())
    public_by_private = {str(j.private_name): name for name, j in state.jails.items()}
    private_by_public = {name: str(j.private_name) for name, j in state.jails.items()}

    monit_by_jail = _fetch_monit_for_jails(raw_jails, public_by_private, ssh_kw)

    jail_rows: list[JailRow] = []
    for j in raw_jails:
        private_name = j["private_name"]
        public_name = public_by_private.get(private_name, private_name)
        managed = private_name in public_by_private

        if managed:
            jail_state = state.jails[public_name]
            ports = [f"{f.proto}/{f.host} → {f.jail}" for f in jail_state.forwards.values()]
            mounts = [f"{short_path(m.host)} → {m.jail}" for m in jail_state.mounts.values()]
        else:
            ports = []
            mounts = []

        ipv4 = [ip for ip in j["ipv4"] if ip != "-"]
        ipv6 = [ip for ip in j["ipv6"] if ip != "-"]

        row = JailRow(
            name=public_name,
            state=j["state"],
            ips=[*ipv4, *ipv6],
            managed=managed,
            ports=ports,
            mounts=mounts,
        )
        if public_name in monit_by_jail:
            row["monit"] = monit_by_jail[public_name]

        jail_rows.append(row)

    jail_private_names = {j["private_name"] for j in raw_jails}
    for public_name in sorted(managed_names):
        private_name = private_by_public[public_name]
        if private_name in jail_private_names:
            continue

        jail = state.jails[public_name]
        ports = [f"{f.proto}/{f.host} → {f.jail}" for f in jail.forwards.values()]
        mounts = [f"{short_path(m.host)} → {m.jail}" for m in jail.mounts.values()]

        jail_rows.append(
            JailRow(
                name=public_name,
                state="Missing",
                ips=[jail.ip] if jail.ip else [],
                managed=True,
                stale=True,
                ports=ports,
                mounts=mounts,
            )
        )

    return StatusInfo(
        pid=pid,
        ssh_host=ssh_kw["ssh_host"],
        ssh_port=ssh_kw["ssh_port"],
        uptime=uptime,
        disk_free=disk_stats["disk_free"],
        disk_total=disk_stats["disk_total"],
        mem_total=mem_stats["mem_total"],
        mem_usable=mem_stats["mem_usable"],
        jail_rows=jail_rows,
    )



from dataclasses import dataclass, field
from pathlib import Path
from typing import Literal

import typer

from jailrun.ansible import run_playbook
from jailrun.cmd.stop import stop_vm
from jailrun.config import (
    derive_plan,
    derive_qemu_fwds,
    derive_qemu_shares,
    diff_jail,
    needs_qemu_restart,
    parse_config,
    resolve_jail,
    resolve_jail_dependencies,
    save_state,
)
from jailrun.misc import lock
from jailrun.network import get_ssh_kw, resolve_jail_ips, resolve_ssh_port, wait_for_ssh
from jailrun.qemu import QemuMode, launch_vm, vm_is_running
from jailrun.remote import fetch_remote_playbook
from jailrun.schemas import ALL_FLAGS, Capability, ChangeFlag, JailConfig, LocalSetupStep, RemoteSetupStep, State
from jailrun.settings import Settings
from jailrun.ui import err, info, ok, warn


@dataclass(frozen=True)
class PlaybookRule:
    playbook: str
    plan: Literal["single", "cumulative"]
    triggers: frozenset[ChangeFlag] = field(default_factory=frozenset)
    requires: frozenset[Capability] = field(default_factory=frozenset)


JAIL_RULES: tuple[PlaybookRule, ...] = (
    PlaybookRule(playbook="jail-create.yml", plan="single", triggers=frozenset({ChangeFlag.CREATE})),
    PlaybookRule(playbook="jail-mounts.yml", plan="single", triggers=frozenset({ChangeFlag.MOUNTS})),
    PlaybookRule(playbook="jail-start.yml", plan="single", triggers=frozenset({ChangeFlag.EXECS})),
    PlaybookRule(
        playbook="jail-dns.yml",
        plan="cumulative",
        triggers=frozenset({ChangeFlag.FORWARDS}),
    ),
    PlaybookRule(playbook="jail-forwards.yml", plan="cumulative", triggers=frozenset({ChangeFlag.FORWARDS})),
    PlaybookRule(
        playbook="jail-yggdrasil.yml",
        plan="single",
        triggers=frozenset({ChangeFlag.CREATE}),
        requires=frozenset({Capability.MESH}),
    ),
    PlaybookRule(
        playbook="jail-pf.yml",
        plan="single",
        triggers=frozenset({ChangeFlag.CREATE}),
        requires=frozenset({Capability.MESH}),
    ),
    PlaybookRule(
        playbook="jail-monit.yml",
        plan="single",
        triggers=frozenset({ChangeFlag.EXECS}),
        requires=frozenset({Capability.EXECS}),
    ),
)


def should_run(rule: PlaybookRule, *, flags: set[ChangeFlag], capabilities: set[Capability]) -> bool:
    if not (rule.triggers & flags):
        return False

    return rule.requires <= capabilities


def run_provisioning(
    jail_cfg: JailConfig,
    jail_name: str,
    jail_ip: str | None,
    settings: Settings,
    state: State,
) -> None:
    for step in jail_cfg.setup.values():
        if step.type != "ansible":
            continue

        if isinstance(step, RemoteSetupStep):
            playbook_path = fetch_remote_playbook(
                step.url,
                cache_dir=settings.playbook_cache_dir,
            )
            run_playbook(
                str(playbook_path),
                jail_name=jail_name,
                jail_ip=jail_ip,
                extra_vars=step.vars or None,
                settings=settings,
                state=state,
            )
        elif isinstance(step, LocalSetupStep):
            run_playbook(
                step.file,
                jail_name=jail_name,
                jail_ip=jail_ip,
                extra_vars=step.vars or None,
                settings=settings,
                state=state,
            )


def up(
    config: Path,
    state: State,
    settings: Settings,
    *,
    names: list[str] | None = None,
    provision: bool = False,
) -> None:
    with lock(settings.state_file):
        _up(
            config=config,
            state=state,
            settings=settings,
            names=names,
            provision=provision,
        )


def _up(
    config: Path,
    state: State,
    settings: Settings,
    *,
    names: list[str] | None = None,
    provision: bool = False,
) -> None:
    alive, _ = vm_is_running(settings.pid_file)
    if not alive:
        err("VM is not running. Run 'jrun start' first.")
        raise typer.Exit(1)

    cfg = parse_config(config)

    if not cfg.jail:
        warn("No jails defined in config.")
        raise typer.Exit(1)

    targets = set(names) if names else set(cfg.jail.keys())
    unknown = targets - set(cfg.jail.keys())

    if unknown:
        err(f"Not in config: {', '.join(sorted(unknown))}")
        raise typer.Exit(1)

    new_state = state.model_copy(deep=True)

    ordered_jails = resolve_jail_dependencies(targets, cfg.jail)

    for name in ordered_jails:
        new_state.jails[name] = resolve_jail(
            jail_config=cfg.jail[name],
            config_base=config.parent.resolve(),
            default_release=f"{settings.bsd_version}-{settings.bsd_release_tag}",
        )

    if needs_qemu_restart(old_state=state, new_state=new_state):
        warn("QEMU wiring changed — restarting VM…")
        stop_vm(settings)

        new_state.ssh_port = resolve_ssh_port(state=new_state, settings=settings)
        new_state.launched_fwds = derive_qemu_fwds(new_state)
        new_state.launched_shares = derive_qemu_shares(new_state)

        launch_vm(state=new_state, mode=QemuMode.SERVER, settings=settings)
        save_state(state=new_state, state_file=settings.state_file)

    ssh_kw = get_ssh_kw(settings, new_state)
    wait_for_ssh(ssh_kw)

    resolve_jail_ips(old_state=state, new_state=new_state, ssh_kw=ssh_kw)
    save_state(state=new_state, state_file=settings.state_file)

    plan = derive_plan(state, new_state)

    run_playbook("jail-stale.yml", plan=plan, settings=settings, state=new_state)
    run_playbook("vm-mounts.yml", plan=plan, settings=settings, state=new_state)

    peers_data = [p.model_dump() for p in new_state.peers]

    provisioned_names = {n for n in new_state.jails if n not in targets}

    deployed: list[str] = []

    for name in ordered_jails:
        jail_state = new_state.jails[name]
        jail_cfg = cfg.jail[name]

        changes = set(ALL_FLAGS) if provision else diff_jail(state.jails.get(name), jail_state)

        if not changes:
            info(f"Jail '{name}' unchanged — skipping.")
            provisioned_names.add(name)
            continue

        provisioned_names.add(name)

        single_plan = plan.for_jail(name)
        cumulative_plan = plan.for_jails(provisioned_names)

        plans = {"single": single_plan, "cumulative": cumulative_plan}

        capabilities: set[Capability] = set()
        if settings.mesh_network:
            capabilities.add(Capability.MESH)
        if single_plan.execs:
            capabilities.add(Capability.EXECS)

        for rule in JAIL_RULES:
            if not should_run(rule, flags=changes, capabilities=capabilities):
                continue

            extra_vars = {}
            if Capability.MESH in rule.requires:
                extra_vars.update({"peers": peers_data})

            run_playbook(
                rule.playbook,
                plan=plans[rule.plan],
                extra_vars=extra_vars,
                settings=settings,
                state=new_state,
            )

        if ChangeFlag.SETUP in changes:
            run_provisioning(
                jail_cfg=jail_cfg,
                jail_name=name,
                jail_ip=jail_cfg.ip or jail_state.ip,
                settings=settings,
                state=new_state,
            )

        save_state(state=new_state, state_file=settings.state_file)
        deployed.append(name)

    if deployed:
        ok(f"Deploy complete ({', '.join(deployed)}).")
    else:
        ok("Nothing to deploy — all jails up to date.")



from enum import StrEnum


class ChangeFlag(StrEnum):
    CREATE = "create"
    MOUNTS = "mounts"
    FORWARDS = "forwards"
    EXECS = "execs"
    SETUP = "setup"


ALL_FLAGS = frozenset(ChangeFlag)


class Capability(StrEnum):
    MESH = "mesh"
    EXECS = "execs"



from jailrun.schemas.base import BaseForwardConfig as BaseForwardConfig
from jailrun.schemas.base import BaseMountConfig as BaseMountConfig
from jailrun.schemas.base import ExecConfig as ExecConfig
from jailrun.schemas.base import HealthcheckConfig as HealthcheckConfig
from jailrun.schemas.base import JailBaseConfig as JailBaseConfig
from jailrun.schemas.base import JailForwardConfig as JailForwardConfig
from jailrun.schemas.base import JailMountConfig as JailMountConfig
from jailrun.schemas.base import LocalSetupStep as LocalSetupStep
from jailrun.schemas.base import RemoteSetupStep as RemoteSetupStep
from jailrun.schemas.base import SetupStep as SetupStep
from jailrun.schemas.base import private_jail_name as private_jail_name
from jailrun.schemas.config import BaseConfig as BaseConfig
from jailrun.schemas.config import Config as Config
from jailrun.schemas.config import JailConfig as JailConfig
from jailrun.schemas.enums import ALL_FLAGS as ALL_FLAGS
from jailrun.schemas.enums import Capability as Capability
from jailrun.schemas.enums import ChangeFlag as ChangeFlag
from jailrun.schemas.plan import ExecPlan as ExecPlan
from jailrun.schemas.plan import JailPlan as JailPlan
from jailrun.schemas.plan import MountPlan as MountPlan
from jailrun.schemas.plan import NullfsPlan as NullfsPlan
from jailrun.schemas.plan import Plan as Plan
from jailrun.schemas.plan import RdrPlan as RdrPlan
from jailrun.schemas.plan import StaleJailPlan as StaleJailPlan
from jailrun.schemas.plan import StaleMountPlan as StaleMountPlan
from jailrun.schemas.plan import StaleNullfsPlan as StaleNullfsPlan
from jailrun.schemas.state import BaseState as BaseState
from jailrun.schemas.state import JailState as JailState
from jailrun.schemas.state import PeerJail as PeerJail
from jailrun.schemas.state import PeerState as PeerState
from jailrun.schemas.state import QemuFwd as QemuFwd
from jailrun.schemas.state import QemuShare as QemuShare
from jailrun.schemas.state import State as State



from pydantic import BaseModel, Field, computed_field

from jailrun.schemas.base import (
    BaseForwardConfig,
    BaseMountConfig,
    ExecConfig,
    JailBaseConfig,
    JailForwardConfig,
    JailMountConfig,
    SetupStep,
    private_jail_name,
)


class BaseConfig(BaseModel):
    setup: dict[str, SetupStep] = Field(default_factory=dict)
    forward: dict[str, BaseForwardConfig] = Field(default_factory=dict)
    mount: dict[str, BaseMountConfig] = Field(default_factory=dict)


class JailConfig(BaseModel):
    name: str
    release: str | None = None
    ip: str | None = None
    base: JailBaseConfig | None = None
    depends: list[str] = Field(default_factory=list)
    forward: dict[str, JailForwardConfig] = Field(default_factory=dict)
    mount: dict[str, JailMountConfig] = Field(default_factory=dict)
    setup: dict[str, SetupStep] = Field(default_factory=dict)
    exec: dict[str, ExecConfig] = Field(default_factory=dict)

    @computed_field
    def private_name(self) -> str:
        return private_jail_name(self.name)


class Config(BaseModel):
    base: BaseConfig | None = None
    jail: dict[str, JailConfig] = Field(default_factory=dict)



from typing import Literal

from pydantic import BaseModel, Field, computed_field

from jailrun.schemas.base import HealthcheckConfig, JailBaseConfig, private_jail_name


class JailPlan(BaseModel):
    name: str
    release: str
    ip: str | None = None
    base: JailBaseConfig | None = None

    @computed_field
    def private_name(self) -> str:
        return private_jail_name(self.name)


class MountPlan(BaseModel):
    mount_tag: str
    target: str


class ExecPlan(BaseModel):
    name: str
    jail: str
    cmd: str
    dir: str = "/"
    env: dict[str, str] = Field(default_factory=dict)
    healthcheck: HealthcheckConfig | None = None

    @computed_field
    def jail_private_name(self) -> str:
        return private_jail_name(self.jail)


class RdrPlan(BaseModel):
    jail: str
    proto: Literal["tcp", "udp"] = "tcp"
    target_port: int
    jail_port: int

    @computed_field
    def jail_private_name(self) -> str:
        return private_jail_name(self.jail)


class NullfsPlan(BaseModel):
    jail: str
    target_path: str
    jail_path: str

    @computed_field
    def jail_private_name(self) -> str:
        return private_jail_name(self.jail)


class StaleMountPlan(BaseModel):
    mount_tag: str
    target: str


class StaleNullfsPlan(BaseModel):
    jail: str
    target_path: str
    jail_path: str

    @computed_field
    def jail_private_name(self) -> str:
        return private_jail_name(self.jail)


class StaleJailPlan(BaseModel):
    name: str

    @computed_field
    def private_name(self) -> str:
        return private_jail_name(self.name)


class Plan(BaseModel):
    jails: list[JailPlan] = Field(default_factory=list)
    mounts: list[MountPlan] = Field(default_factory=list)
    execs: list[ExecPlan] = Field(default_factory=list)
    jail_rdrs: list[RdrPlan] = Field(default_factory=list)
    jail_mounts: list[NullfsPlan] = Field(default_factory=list)
    stale_jails: list[StaleJailPlan] = Field(default_factory=list)
    stale_mounts: list[StaleMountPlan] = Field(default_factory=list)
    stale_jail_mounts: list[StaleNullfsPlan] = Field(default_factory=list)

    def for_jail(self, name: str) -> "Plan":
        return Plan(
            jails=[j for j in self.jails if j.name == name],
            jail_mounts=[m for m in self.jail_mounts if m.jail == name],
            stale_jail_mounts=[m for m in self.stale_jail_mounts if m.jail == name],
            jail_rdrs=[r for r in self.jail_rdrs if r.jail == name],
            execs=[e for e in self.execs if e.jail == name],
        )

    def for_jails(self, names: set[str]) -> "Plan":
        return Plan(
            jails=[j for j in self.jails if j.name in names],
            jail_rdrs=[r for r in self.jail_rdrs if r.jail in names],
        )



from typing import Literal

from pydantic import BaseModel, Field, computed_field

from jailrun.schemas.base import (
    BaseForwardConfig,
    BaseMountConfig,
    ExecConfig,
    JailBaseConfig,
    JailForwardConfig,
    JailMountConfig,
    SetupStep,
    private_jail_name,
)


class JailState(BaseModel):
    name: str
    base: JailBaseConfig | None = None
    release: str
    ip: str | None = None
    forwards: dict[str, JailForwardConfig] = Field(default_factory=dict)
    mounts: dict[str, JailMountConfig] = Field(default_factory=dict)
    execs: dict[str, ExecConfig] = Field(default_factory=dict)
    setup: dict[str, SetupStep] = Field(default_factory=dict)

    @computed_field
    def private_name(self) -> str:
        return private_jail_name(self.name)


class BaseState(BaseModel):
    setup: dict[str, SetupStep] = Field(default_factory=dict)
    forwards: dict[str, BaseForwardConfig] = Field(default_factory=dict)
    mounts: dict[str, BaseMountConfig] = Field(default_factory=dict)


class QemuFwd(BaseModel):
    proto: Literal["tcp", "udp"]
    host: int
    guest: int


class QemuShare(BaseModel):
    host: str
    id: str
    mount_tag: str


class PeerJail(BaseModel):
    name: str
    ygg_address: str


class PeerState(BaseModel):
    alias: str
    direction: Literal["init", "joined"]
    paired_at: str
    jails: list[PeerJail] = Field(default_factory=list)


class State(BaseModel):
    version: int = 1
    base: BaseState = Field(default_factory=BaseState)
    jails: dict[str, JailState] = Field(default_factory=dict)
    launched_fwds: list[QemuFwd] = Field(default_factory=list)
    launched_shares: list[QemuShare] = Field(default_factory=list)
    ssh_port: int | None = None
    peers: list[PeerState] = Field(default_factory=list)
